/*
 load a RAW PPM in to the framebuffer
 Works with color bpp 16 

 PPM comment not supported (yet)

 (C) 2002Michele Andreoli
  GPL
*/

#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <linux/fb.h>
#include <sys/mman.h>

#include "resize.c"


#define MAX(x,y) y>x?(y):(x)
#define MIN(x,y) y>x?(x):(y)


typedef unsigned char uchar;

#define num_colors 1 << bpp 
#define SIZE x0 * y0 * cpp 
#define DELAY 1 

int color;
uchar *buffer;
int cpp,bpp;
int x0,y0;  /* screen dimension */
int w,h;  /* image dimension */
uchar *image;	/* the mmapped image buffer */
int X0,Y0,X1,Y1;	/* current area in use */

int fd,fbfd;

    struct fb_var_screeninfo vinfo;
    struct fb_fix_screeninfo finfo;

int pid=0;	// child process pid

/* options */

int opt_fullscreen=0;
int opt_refresh=0;
int opt_wait=0;
int opt_clear=0;
int opt_scale=1;
int opt_blink=0;
int opt_diagonal=0;
int opt_bounce=0;
int opt_timeout=0;

int opt_x=-1;
int opt_y=-1;

/*
 create the 1-byte encoded 16 bit color
*/

inline uchar make8color(uchar r, uchar g,
        uchar b)
{
    return (
        (((r >> 5) & 7) << 5) |
        (((g >> 5) & 7) << 2) |
         ((b >> 6) & 3)       );
}


/*
 create the 2-byte encoded 16 bit color
*/

inline unsigned short make16color(uchar r, uchar g,
 uchar b)
{
    return (
        (((r >> 3) & 31) << 11) |
        (((g >> 2) & 63) << 5)  |
         ((b >> 3) & 31)        );
}


/*
	write an RGB value in the framebuffer
*/

inline void rgb16(int x, int y, uchar r, uchar g, uchar b) {

// solo a 16 bit
unsigned short color;
char *p;

    color=(unsigned short) make16color(r,g,b);
    memcpy(buffer + cpp*(y * x0 + x), &color,2);
}

/*
 rustic pseudocolor implementation (not used here) 
*/

inline void pixel(int x, int y, uchar color ) {
uchar r,g,b;

r=color & 0xE0;
g=color & 0x1C;
b=color & 0x02;

rgb16(x,y,r*36,g*36,b*60);
}

/*
 clear the framebuffer
*/

void clear()
{
int i;
memset(buffer,0,SIZE);
}

/*
  load the PPM in the memory
*/

void load_ppm(char* filename)
{
uchar c[3]; 
int x,y,size;
int i=0;
char dummy[128];
char format[32];

int offx,offy;

FILE *f;


//printf("loading %s %i x %i\n",filename,w,h);

if ( !strcmp(filename,"-") ) {
	f=stdin;
} else {
	f=fopen(filename,"r");
}

fd=fileno(f);

fscanf(f,"%s\n",dummy); 	// skip P6
fscanf(f,"#%s\n",dummy);	// skip comment
fscanf(f,"%i %i\n",&w,&h);	// read width height
fscanf(f,"%i",&size);		// num. colors


fprintf(stderr,"loading %s %i x %i max=%i\n",filename,w,h,size);
image=(uchar*) malloc(w*h*3);

//read(fd,image, w*h*3);
fscanf(f,"%*c");
sprintf(format,"%%%ic",w*h*3);
fscanf(f,format,image);

//printf("first %x %x %x\n", image[0],image[1],image[2]);
close(fd);


}

int scale()
{

if ( opt_fullscreen )  {image=simple_resize(image,w,h,x0,y0); w=x0; h=y0;}
if ( opt_scale > 1 )  {
        image=simple_resize(image,w,h,opt_scale*w,opt_scale*h);
        w*=opt_scale; h*=opt_scale;
}

}


/*
 draw the image, shifting by (offx, offy)
*/


int draw(int offx, int offy) {
int x,y;
int i=0;


// calculate offsets in order to center the image
if ( (offx == -1) && (offy == -1 ) ) {
	offx=(x0-w)/2;;
	offy=(y0-h)/2;
}

// store the work area geometry
X0=offx;
Y0=offy;
X1=X0+w;
Y1=Y0+h;

// draw the image

for (y=0; y< h; y++) {
        for(x=0; x<w; x++) {
                rgb16(X0+x,Y0+y, image[i+0],image[i+1],image[i+2]);
                i+=3;
        }
}

}

/*
cancel a rettangular area 
*/ 

int blank(int x1,int y1, int x2, int y2)
{
int x,y;
int black=0;

int ox1,ox2,oy1,oy2;

ox1=MIN(x1,x2); ox2=MAX(x1,x2);
oy1=MIN(y1,y2); oy2=MAX(y1,y2);

for (y=oy1; y< oy2; y++) {
        for(x=ox1; x<ox2; x++) {
			memcpy(buffer + cpp*(y*x0+x),&black,2); 
        }
}

}

/*
 cancel the image
*/

int cancel()
{
blank(X0,Y0,X1,Y1);
}

/*
erase only the "exposed" part of the  current image
*/

int clip ( int offx, int offy ) {

if ( offy<=Y0 ) blank(X0,offy+h,X1,Y1); 
if ( offy>Y0 ) blank(X0,Y0,X1,offy); 


if ( offx<=X0 ) blank(offx+w,Y0,X1,Y1); 
if ( offx>X0 ) blank(X0,Y0,offx,Y1); 

}


/*
initialize the framebuffer device
*/


void init()
{

fbfd = open("/dev/fb0", O_RDWR);
   // Get fixed screen information
    if (ioctl(fbfd, FBIOGET_FSCREENINFO, &finfo)) {
        fprintf(stderr,"Error reading fixed information.\n");
        exit(2);
    }

    // Get variable screen information
    if (ioctl(fbfd, FBIOGET_VSCREENINFO, &vinfo)) {
        fprintf(stderr,"Error reading variable information.\n");
        exit(3);
    }


bpp=vinfo.bits_per_pixel;
cpp=bpp/8;
x0=vinfo.xres; y0=vinfo.yres;
buffer = mmap(0, SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, fbfd, 0);

fprintf(stderr,"bpp %d, cpp %d, %d x %d\n",bpp,cpp,x0,y0);

if ( bpp != 16 ) { fprintf(stderr,"bpp!=16\n"); exit(1);}

}

/*
blink demo
*/

int blink() {

        while (1) {
        draw(opt_x,opt_y);
        sleep(1);
        cancel();
        sleep(1);
        }
}


/*
diagonal motion demo
*/

int diagonal() {
int x,y;
int v=opt_diagonal;
for (x=0; (x+opt_scale*w<x0) && (x+opt_scale*h<y0) ; x+=v) {
	draw(x,x); 
	clip(x+v,x+v);
	}

}


/*
 read from /proc/interrupts the number of keyboard's interrupts event
*/

int get_interrupts()
{
int irq,n;
char *dummy[120];
FILE *f;

f=fopen("/proc/interrupts","r");

fscanf(f,"%[^\n]\n",dummy);         // skip 1th row
fscanf(f,"%[^\n]\n",dummy);         // skip 1th row

fscanf(f,"%3i:%11i",&irq,&n);      // read irq and #interrupt
//printf("irq %i, inter %i\n", irq,n);
fclose(f);
return n;
}


int timeout()
{
int irq,current,previous,n;
int installed=0;
int t=0;
int running=0;
int rc;

// initialize interrupts counter

previous=get_interrupts();
printf("init irq %i\n",previous);

while(1) {
	sleep(1);
	current=get_interrupts(); // current irqs
	printf("figlio %i, t=%i, installed=%i,running=%i\n",pid,t,installed,running);
	printf("current irq %i\n",current);

	if (installed) t++;

	if ( (t > opt_timeout) && ! running ) {
		running=1; t=0;
		if ( pid=fork() ) {
		//padre
		} else {
		bounce(1);
		}
	}
	
	if ( current == previous ) {
		if (!installed) {
		printf("keyboard handler enabled\n");
		installed=1;
		}
	} else {
		printf("keyboard handler disabled\n");
		installed=0; t=0;
		if (running && pid ) { running=0; kill(pid, SIGTERM);
			waitpid(pid,&rc,0);
			 }
	}
	
	previous=current;

} // end while
} // end proc



/*
bouncing motion demo
*/

int bounce(int v) {
int x=0,y=0;
int nx=0,ny=0;
int vx=v;
int vy=v;
int safe=0;

printf("bouncing demo: %i\n",v);

while ( 1 ) {
	draw(x,y);
	if ( (X0+vx +safe >x0) || (X0+vx -safe < 0) ) vx*=-1;
	if ( (Y0+vx +safe >y0) || (Y0+vy -safe < 0) ) vy*=-1;

	if ( (X1+vx +safe >x0) || (X1+vx -safe< 0) ) vx*=-1;
	if ( (Y1+vx +safe >y0) || (Y1+vy -safe < 0) ) vy*=-1;

	clip(x+vx,y+vy);
	x+=vx; y+=vy;
}



}





int usage(char *program) {

fprintf(stderr,"%s, Display (rawbits) PPM on Linux framebuffer (C) 2002 M. Andreoli\n",program);
fprintf(stderr, "Usage: %s [opts] ppmfile|-\n",program); 
fprintf(stderr, "Options:\n"); 
fprintf(stderr, "\t -x offx -y offy (XY offsets, up-left corner) \n"); 
fprintf(stderr, "\t -c (clear the framebuffer, at the end) \n"); 
fprintf(stderr, "\t -r (refresh mode) \n"); 
fprintf(stderr, "\t -f (fullscreen) \n"); 
fprintf(stderr, "\t -w [secs]; wait before exit; -w -1 means until keypressed\n"); 
fprintf(stderr, "\t -s [integer factor] (scale)\n"); 
fprintf(stderr, "\t -b (blink demo)\n"); 
fprintf(stderr, "\t -d velocity|0 (diagonal demo)\n"); 
fprintf(stderr, "\t -B velocity|0 (bouncing demo)\n"); 
fprintf(stderr, "\t -t timeout (screen saver mode)\n"); 


exit(1);
}

void quit()
{
free(image);
if (opt_clear) clear();
munmap(buffer, SIZE);
close(fbfd);
exit(0);
}


int main(int argc, char** argv)
{
int i;
int d;
int k=0;
char *filename;
int width,height;
int opt;

/* install signal handler */

signal(SIGTERM,quit);
signal(SIGALRM,quit);

while ((opt = getopt(argc, argv, "bfcrw:s:x:y:d:B:t:")) != EOF)    {

switch( opt ) {
case 'f': opt_fullscreen=1; break;
case 'r': opt_refresh=1; break;
case 'w': opt_wait=atoi(optarg); break;
case 's': opt_scale=atoi(optarg); break;
case 'x': opt_x=atoi(optarg); break;
case 'y': opt_y=atoi(optarg); break;
case 'd': opt_diagonal=atoi(optarg); break;
case 'B': opt_bounce=atoi(optarg); break;
case 't': opt_timeout=atoi(optarg); break;
case 'c': opt_clear=1;break;
case 'b': opt_blink=1;break;
} 

}

filename=argv[optind];
if (filename==0 ) usage(argv[0]);

if (opt_wait>0 ) {
alarm(0);
alarm(opt_wait);
}

init();	// init framebuffer
clear();
load_ppm(filename);
scale();

if( opt_blink) blink() ; 
if( opt_diagonal) {diagonal(); goto finish;} 
if( opt_bounce) bounce(opt_bounce); 
if( opt_timeout) timeout(); 

if ( opt_refresh  ) { while(1) draw(opt_x,opt_y);}
else
	draw(opt_x,opt_y);


finish:
if (opt_wait==-1) getchar();
if (opt_wait>0 ) sleep(opt_wait);
quit();
}
